#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from .preshared_key import PresharedKey
from .public_key import PublicKey

class WireGuardPeer(object):
    __slots__ = (
        "public_key",
        "preshared_key",
        "endpoint",
        "persistent_keepalive_interval",
        "last_handshake_time",
        "rx_bytes",
        "tx_bytes",
        "allowedips",
        "protocol_version",
    )

    def __init__(self, peer, spill_preshared_keys=False):
        self.public_key = PublicKey(peer.get_attr('WGPEER_A_PUBLIC_KEY'))
        self.preshared_key = peer.get_attr('WGPEER_A_PRESHARED_KEY')
        if not spill_preshared_keys:
            self.preshared_key = self.preshared_key is not None
        else:
            self.preshared_key = PresharedKey(self.preshared_key)
        self.endpoint = peer.get_attr('WGPEER_A_ENDPOINT')
        self.persistent_keepalive_interval = peer.get_attr('WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL')
        self.last_handshake_time = peer.get_attr('WGPEER_A_LAST_HANDSHAKE_TIME')
        if not self.last_handshake_time > 0:
            self.last_handshake_time = None
        self.rx_bytes = peer.get_attr('WGPEER_A_RX_BYTES')
        self.tx_bytes = peer.get_attr('WGPEER_A_TX_BYTES')
        self.protocol_version = peer.get_attr('WGPEER_A_PROTOCOL_VERSION')

        self.allowedips = []
        for allowedip in peer.get_attr('WGPEER_A_ALLOWEDIPS') or []:
            self.allowedips.append(allowedip.network())

    def __repr__(self):
        return repr({
            'public_key': self.public_key,
            'preshared_key': self.preshared_key,
            'endpoint': self.endpoint,
            'persistent_keepalive_interval': self.persistent_keepalive_interval,
            'last_handshake_time': self.last_handshake_time,
            'rx_bytes': self.rx_bytes,
            'tx_bytes': self.tx_bytes,
            'allowedips': self.allowedips,
            'protocol_version': self.protocol_version,
        })

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from base64 import b64encode, b64decode

class Key:
    __slots__ = ('__value')

    def __init__(self, key=bytes(32)):
        if isinstance(key, Key):
            self.__value = key.__value
        elif isinstance(key, bytes):
            self.__value = key
        elif isinstance(key, bytearray):
            self.__value = bytes(key)
        elif isinstance(key, str):
            self.__value = b64decode(key)
        else:
            raise TypeError()

        assert isinstance(self.__value, bytes)
        assert 32 == len(self.__value)

    def __str__(self):
        return b64encode(self.__value).decode('utf-8')

    def __bytes__(self):
        return self.__value

    def __repr__(self):
        return f'{type(self).__name__}({repr(str(self))})'

    def __bool__(self):
        return self.__value != bytes(32)

    def __eq__(self, other):
        if isinstance(other, Key):
            return self.__value == other.__value
        elif isinstance(other, (bytes, bytearray)):
            return self.__value == other
        elif isinstance(other, str):
            return self.__value == b64decode(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.__value)

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from .key import Key

from base64 import b64decode
from hashlib import shake_128, blake2s
from ipaddress import ip_network, IPv4Network, IPv6Network
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PublicKey

class PublicKey(Key):
    def __init__(self, key=None):
        if key is None:
            super().__init__()
        elif isinstance(key, PublicKey):
            super().__init__(key)
        elif isinstance(key, X25519PublicKey):
            super().__init__(key.public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            ))
        elif isinstance(key, (bytes, bytearray)):
            super().__init__(X25519PublicKey.from_public_bytes(key).public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            ))
        elif isinstance(key, str):
            super().__init__(X25519PublicKey.from_public_bytes(b64decode(key)).public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            ))
        else:
            raise TypeError("key must be PublicKey, bytes, bytearray, or str")

    def __eq__(self, other):
        if isinstance(other, Key) and not isinstance(other, PublicKey):
            return NotImplemented

        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def orchid(self, secret=b'', network=None):
        if network is None:
            return self.orchid6(secret)

        if isinstance(secret, str):
            secret = secret.encode('utf-8')
        elif not isinstance(secret, (bytes, bytearray)):
            secret = bytes(secret)

        if not isinstance(network, (IPv4Network, IPv6Network, )):
            network = ip_network(network)

        hash = shake_128(secret + bytes(self)).digest(network.max_prefixlen//8)
        mask = int.from_bytes(network.hostmask.packed, byteorder='big')
        host = int.from_bytes(hash, byteorder='big')
        addr = network[host & mask]

        if addr == network.network_address:
            addr += 1
        elif addr == network.broadcast_address:
            addr -= 1

        assert addr != network.network_address, "Generated network address"
        assert addr != network.broadcast_address, "Generated broadcast address"
        assert addr in network, "Generated out-of-network address"
        return addr

    def orchid4(self, secret=b''):
        return self.orchid(secret, IPv4Network("100.64.0.0/10"))

    def orchid6(self, secret=b''):
        return self.orchid(secret, IPv6Network("2001:20::/28"))


    def lla(self, secret=b'', network=None):
        if network is None:
            return self.lla6(secret)

        if isinstance(secret, str):
            secret = secret.encode('utf-8')
        elif not isinstance(secret, (bytes, bytearray)):
            secret = bytes(secret)

        if not isinstance(network, (IPv4Network, IPv6Network, )):
            network = ip_network(network)

        hash = blake2s(bytes(self),
            digest_size=32,
            key=secret,
            ).digest()[:network.max_prefixlen//8]
        mask = int.from_bytes(network.hostmask.packed, byteorder='big')
        host = int.from_bytes(hash, byteorder='big')
        addr = network[host & mask]

        if addr == network.network_address:
            addr += 1
        elif addr == network.broadcast_address:
            addr -= 1

        assert addr != network.network_address, "Generated network address"
        assert addr != network.broadcast_address, "Generated broadcast address"
        assert addr in network, "Generated out-of-network address"
        return addr

    def lla4(self, secret=b''):
        return self.lla(secret, IPv4Network("169.254.0.0/16"))

    def lla6(self, secret=b''):
        return self.lla(secret, IPv6Network("fe80::/10"))

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from .private_key import PrivateKey
from .public_key import PublicKey
from .wireguardpeer import WireGuardPeer

class WireGuardInfo(object):
    __slots__ = (
        "ifindex",
        "ifname",
        "private_key",
        "public_key",
        "listen_port",
        "fwmark",
        "peers",
    )

    def __init__(self, messages, spill_private_key, spill_preshared_keys):
        self.ifindex = messages[0].get_attr('WGDEVICE_A_IFINDEX')
        self.ifname = messages[0].get_attr('WGDEVICE_A_IFNAME')
        self.private_key = messages[0].get_attr('WGDEVICE_A_PRIVATE_KEY')
        if not spill_private_key:
            self.private_key = self.private_key is not None
        elif self.private_key is not None:
            self.private_key = PrivateKey(self.private_key)
        self.public_key = messages[0].get_attr('WGDEVICE_A_PUBLIC_KEY')
        if self.public_key is not None:
            self.public_key = PublicKey(self.public_key)
        self.listen_port = messages[0].get_attr('WGDEVICE_A_LISTEN_PORT')
        self.fwmark = messages[0].get_attr('WGDEVICE_A_FWMARK')

        self.peers = { }

        for message in messages:
            wgp = lambda p: WireGuardPeer(p, spill_preshared_keys)
            for peer in map(wgp, message.get_attr('WGDEVICE_A_PEERS') or []):
                assert peer.public_key not in self.peers
                self.peers[peer.public_key] = peer

    def __repr__(self):
        return repr({
            'ifindex': self.ifindex,
            'ifname': self.ifname,
            'private_key': self.private_key,
            'public_key': self.public_key,
            'listen_port': self.listen_port,
            'fwmark': self.fwmark,
            'peers': self.peers,
        })

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from .key import Key

from secrets import token_bytes

class PresharedKey(Key):
    def __init__(self, key=None):
        if isinstance(key, type(None)):
            super().__init__()
        elif isinstance(key, (type(None), PresharedKey, bytes, bytearray, str)):
            super().__init__(key)
        else:
            raise TypeError("key must be PresharedKey, bytes, bytearray, or str")

    def __eq__(self, other):
        if isinstance(other, Key) and not isinstance(other, PresharedKey):
            return NotImplemented

        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()

    @staticmethod
    def generate():
        return PresharedKey(token_bytes(32))

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from .key import Key
from .public_key import PublicKey

from base64 import b64decode
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey

class PrivateKey(Key):
    def __init__(self, key=None):
        if key is None:
            super().__init__()
        elif isinstance(key, PrivateKey):
            super().__init__(key)
        elif isinstance(key, X25519PrivateKey):
            super().__init__(key.private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption(),
            ))
        elif isinstance(key, (bytes, bytearray)):
            super().__init__(X25519PrivateKey.from_private_bytes(key).private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption(),
            ))
        elif isinstance(key, str):
            super().__init__(X25519PrivateKey.from_private_bytes(b64decode(key)).private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption(),
            ))
        else:
            raise TypeError("key must be PrivateKey, bytes, bytearray, or str")

    def __eq__(self, other):
        if isinstance(other, Key) and not isinstance(other, PrivateKey):
            return NotImplemented

        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def public_key(self):
        return PublicKey(X25519PrivateKey.from_private_bytes(bytes(self)).public_key().public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
        ))

        if isinstance(other, Key) and not isinstance(other, PrivateKey):
            return NotImplemented

        return super().__eq__(other)

    @staticmethod
    def generate():
        return PrivateKey(X25519PrivateKey.generate().private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption(),
        ))

#



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from pyroute2 import netlink

from .preshared_key import PresharedKey
from .private_key import PrivateKey
from .public_key import PublicKey
from .wireguardinfo import WireGuardInfo

class WireGuard(object):
    __slots__ = ( "__socket" )

    from .nlas import device as __device

    def __init__(self, **kwargs):
        self.__socket = netlink.generic.GenericNetlinkSocket()
        self.__socket.bind('wireguard', self.__device)

    def __del__(self):
        self.__socket.close()

    def __get(self, device):
        flags = netlink.NLM_F_REQUEST | netlink.NLM_F_DUMP
        return self.__socket.nlm_request(device, msg_type=self.__socket.prid, msg_flags=flags)

    def __set(self, device):
        flags = netlink.NLM_F_ACK | netlink.NLM_F_REQUEST
        return self.__socket.nlm_request(device, msg_type=self.__socket.prid, msg_flags=flags)

    def get_interface(self, interface, spill_private_key=False, spill_preshared_keys=False):
        device = self.__device.get_device(interface)
        messages = self.__get(device)

        return WireGuardInfo(messages, spill_private_key, spill_preshared_keys)

    def set_interface(self, interface,
            private_key=None,
            listen_port=None,
            fwmark=None,
            replace_peers=False,
            ):

        device = self.__device.set_device(interface)

        if replace_peers:
            device['attrs'].append(('WGDEVICE_A_FLAGS', device.flag.REPLACE_PEERS.value))

        if private_key is not None:
            if isinstance(private_key, PrivateKey):
                private_key = bytes(private_key)
            elif not isinstance(private_key, (bytes, bytearray)):
                private_key = bytes(PrivateKey(private_key))
            device['attrs'].append(('WGDEVICE_A_PRIVATE_KEY', private_key))

        if listen_port is not None:
            device['attrs'].append(('WGDEVICE_A_LISTEN_PORT', listen_port))

        if fwmark is not None:
            device['attrs'].append(('WGDEVICE_A_FWMARK', fwmark))

        return self.__set(device)

    def remove_peers(self, interface, *public_keys):
        device = self.__device.set_device(interface)
        device['attrs'].append(('WGDEVICE_A_PEERS', []))

        for public_key in public_keys:
            peer = self.__device.peer()
            if isinstance(public_key, PublicKey):
                public_key = bytes(public_key)
            elif not isinstance(public_key, (bytes, bytearray)):
                public_key = bytes(PublicKey(public_key))
            peer['attrs'].append(('WGPEER_A_PUBLIC_KEY', public_key))
            peer['attrs'].append(('WGPEER_A_FLAGS', peer.flag.REMOVE_ME.value))
            device.get_attr('WGDEVICE_A_PEERS').append({'attrs': peer['attrs']})

        return self.__set(device)

    def set_peer(self, interface, public_key,
            preshared_key=None,
            endpoint=None,
            persistent_keepalive_interval=None,
            allowedips=None,
            replace_allowedips=None,
            update_only=False,
            ):

        device = self.__device.set_device(interface)

        peer = device.peer()
        if isinstance(public_key, PublicKey):
            public_key = bytes(public_key)
        elif not isinstance(public_key, (bytes, bytearray)):
            public_key = bytes(PublicKey(public_key))
        peer['attrs'].append(('WGPEER_A_PUBLIC_KEY', public_key))

        if replace_allowedips is None and allowedips is not None:
            replace_allowedips = True

        flags = 0
        if replace_allowedips:
            flags |= peer.flag.REPLACE_ALLOWEDIPS.value
        if update_only:
            flags |= peer.flag.UPDATE_ONLY.value
        if flags:
            peer['attrs'].append(('WGPEER_A_FLAGS', flags))

        if preshared_key is not None:
            if isinstance(preshared_key, PresharedKey):
                preshared_key = bytes(preshared_key)
            elif not isinstance(preshared_key, (bytes, bytearray)):
                preshared_key = bytes(PresharedKey(preshared_key))
            peer['attrs'].append(('WGPEER_A_PRESHARED_KEY', preshared_key))

        if endpoint is not None:
            peer['attrs'].append(('WGPEER_A_ENDPOINT', self.__device.peer.sockaddr.frob(endpoint)))

        if persistent_keepalive_interval is not None:
            peer['attrs'].append(('WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL', persistent_keepalive_interval))

        if allowedips is not None:
            peer['attrs'].append(('WGPEER_A_ALLOWEDIPS', []))

            for allowedip in allowedips:
                peer.get_attr('WGPEER_A_ALLOWEDIPS').append({'attrs': self.__device.peer.allowedip.frob(allowedip)['attrs']})

        device['attrs'].append(('WGDEVICE_A_PEERS', [{'attrs': peer['attrs']}]))
        return self.__set(device)

    def replace_allowedips(self, interface, *public_keys):
        device = self.__device.set_device(interface)
        device['attrs'].append(('WGDEVICE_A_PEERS', []))

        for public_key in public_keys:
            peer = self.__device.peer()
            if isinstance(public_key, PublicKey):
                public_key = bytes(public_key)
            elif not isinstance(public_key, (bytes, bytearray)):
                public_key = bytes(PublicKey(public_key))
            peer['attrs'].append(('WGPEER_A_PUBLIC_KEY', public_key))
            peer['attrs'].append(('WGPEER_A_FLAGS', peer.flag.REPLACE_ALLOWEDIPS.value))
            device.get_attr('WGDEVICE_A_PEERS').append({'attrs': peer['attrs']})

        return self.__set(device)

#
